# Copyright 2019-2022 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
import pickle
from pathlib import Path

import pandas as pd

from asreview._version import get_versions
from asreview.data import ASReviewData


class CacheDataError(Exception):
    pass


def _get_cache_data_path(fp_data):

    return Path(fp_data) \
        .with_suffix(fp_data.suffix + ".pickle")


def _read_data_from_cache(fp_data, version_check=True):

    fp_data_pickle = _get_cache_data_path(fp_data)

    try:
        # get the pickle data
        with open(fp_data_pickle, 'rb') as f_pickle_read:
            data_obj, data_obj_version = pickle.load(f_pickle_read)

        # validate data object
        if not isinstance(data_obj.df, pd.DataFrame):
            raise ValueError()

        # drop cache files generated by older versions
        if (not version_check) or (get_versions()['version'] == data_obj_version):
            return data_obj

    except FileNotFoundError:
        # file not available
        pass
    except Exception as err:
        # problem loading pickle file or outdated
        # remove the pickle file
        logging.error(f"Error reading cache file: {err}")
        try:
            os.remove(fp_data_pickle)
        except FileNotFoundError:
            pass

    raise CacheDataError()


def _write_data_to_cache(fp_data, data_obj):

    fp_data_pickle = _get_cache_data_path(fp_data)

    logging.info("Store a copy of the data in a pickle file.")
    with open(fp_data_pickle, 'wb') as f_pickle:
        pickle.dump((data_obj, get_versions()['version']), f_pickle)


def read_data(project, use_cache=True, save_cache=True):
    """Get ASReviewData object from file.

    Parameters
    ----------
    project_path: str, iterable
        The project identifier.
    use_cache: bool
        Use the pickle file if available.
    save_cache: bool
        Save the file to a pickle file if not available.

    Returns
    -------
    ASReviewData:
        The data object for internal use in ASReview.

    """

    try:
        fp_data = Path(
            project.project_path, "data", project.config["dataset_path"])
    except Exception:
        raise FileNotFoundError("Dataset not found")

    # use cache file
    if use_cache:
        try:
            return _read_data_from_cache(fp_data)
        except CacheDataError:
            pass

    # load from file
    data_obj = ASReviewData.from_file(fp_data)

    # save a pickle version
    if save_cache:
        _write_data_to_cache(fp_data, data_obj)

    return data_obj



# Copyright 2019-2022 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Command Line Interface (CLI) for ASReview project."""
import argparse
import logging
import sys

from asreview import __version__
from asreview.utils import get_entry_points

PROG_DESCRIPTION = "Automated Systematic Review (ASReview)."

# Internal or deprecated entry points. These entry points
# are not displayed in the help page of the  user interface.
INTERNAL_ENTRY_POINTS = ["web_run_model"]
DEPRECATED_ENTRY_POINTS = ["oracle"]


def _sort_entry_points(entry_points):

    entry_points_copy = entry_points.copy()

    entry_points_sorted = {
        "lab": entry_points_copy.pop("lab"),
        "simulate": entry_points_copy.pop("simulate")
    }

    entry_points_sorted.update(entry_points_copy)

    return entry_points_sorted


def _output_available_entry_points(entry_points):

    entry_points_sorted = _sort_entry_points(entry_points)

    description_list = []
    for name, entry in entry_points_sorted.items():

        # don't display the internal entry points
        if name in INTERNAL_ENTRY_POINTS:
            continue

        # don't display the deprecated entry points
        if name in DEPRECATED_ENTRY_POINTS:
            continue

        # try to load entry points, hide when failing on loading
        try:
            description_list.append(entry.load()().format(name))
        except ModuleNotFoundError:
            logging.warning(
                f"Plugin with entry point {name} could not be loaded.")
    return "\n\n".join(description_list)


def main():
    # Get the available entry points.
    entry_points = get_entry_points("asreview.entry_points")

    # Try to load the entry point if available.
    if len(sys.argv) > 1 and sys.argv[1] in entry_points:

        try:
            entry = entry_points[sys.argv[1]]
            entry.load()().execute(sys.argv[2:])
        except ModuleNotFoundError:
            raise ValueError(
                f"Plugin with entry point {entry.name} could not be loaded.")

    # Print help message if subcommand not given or incorrect
    else:

        # format the available subcommands
        description_subcommands = _output_available_entry_points(entry_points)

        parser = argparse.ArgumentParser(
            prog="asreview",
            formatter_class=argparse.RawTextHelpFormatter,
            description=PROG_DESCRIPTION
        )
        parser.add_argument(
            "subcommand",
            nargs="?",
            default=None,
            help=f"The subcommand to launch. Available commands:\n\n"
            f"{description_subcommands}"
        )

        # version
        parser.add_argument(
            "-V", "--version",
            action='store_true',
            help="print the ASR version number and exit")

        args, _ = parser.parse_known_args()

        # output the version
        if args.version:
            print(__version__)
            return

        parser.print_help()


# execute main function
if __name__ == "__main__":
    main()



# Copyright 2019-2022 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Simulation entry point and utils."""

import logging
import shutil
from pathlib import Path

from asreview.compat import convert_id_to_idx
from asreview.config import ASCII_LOGO
from asreview.config import DEFAULT_BALANCE_STRATEGY
from asreview.config import DEFAULT_FEATURE_EXTRACTION
from asreview.config import DEFAULT_MODEL
from asreview.config import DEFAULT_N_INSTANCES
from asreview.config import DEFAULT_N_PRIOR_EXCLUDED
from asreview.config import DEFAULT_N_PRIOR_INCLUDED
from asreview.config import DEFAULT_QUERY_STRATEGY
from asreview.config import EMAIL_ADDRESS
from asreview.config import GITHUB_PAGE
from asreview.data import load_data
from asreview.entry_points.base import BaseEntryPoint
from asreview.entry_points.base import _base_parser
from asreview.io.paper_record import preview_record
from asreview.models.balance.utils import get_balance_model
from asreview.models.classifiers import get_classifier
from asreview.models.feature_extraction import get_feature_model
from asreview.models.query import get_query_model
from asreview.project import ASReviewProject
from asreview.project import ProjectExistsError
from asreview.project import open_state
from asreview.review.simulate import ReviewSimulate
from asreview.settings import ASReviewSettings
from asreview.types import type_n_queries
from asreview.utils import get_random_state
from asreview.webapp.io import read_data

ASCII_MSG_SIMULATE = """
---------------------------------------------------------------------------------
|                                                                                |
|  Welcome to ASReview LAB - AI-assisted systematic reviews software.            |
|  In simulation mode the computer will simulate how well ASReview LAB           |
|  could have accelerate the systematic review of your dataset.                  |
|  You can sit back and relax while the computer runs this simulation.           |
|                                                                                |
|  GitHub page:        {0: <58}|
|  Questions/remarks:  {1: <58}|
|                                                                                |
---------------------------------------------------------------------------------
""".format(GITHUB_PAGE, EMAIL_ADDRESS)  # noqa


def _get_dataset_path_from_args(args_dataset):
    """Remove 'benchmark:' from the dataset name and add .csv suffix.

    Parameters
    ----------
    args_dataset : str
        Name of the dataset.

    Returns
    -------
    str
        Dataset name without 'benchmark:' if it started with that,
        and with .csv suffix.
    """
    if args_dataset.startswith('benchmark:'):
        args_dataset = args_dataset[10:]

    return Path(args_dataset).with_suffix('.csv').name


def _set_log_verbosity(verbose):
    if verbose == 0:
        logging.getLogger().setLevel(logging.WARNING)
    elif verbose == 1:
        logging.getLogger().setLevel(logging.INFO)
    elif verbose >= 2:
        logging.getLogger().setLevel(logging.DEBUG)


class SimulateEntryPoint(BaseEntryPoint):
    """Entrypoint for simulation."""

    description = "Simulate the performance of ASReview."

    def execute(self, argv):  # noqa

        # parse arguments
        parser = _simulate_parser()
        args = parser.parse_args(argv)

        # change the verbosity
        _set_log_verbosity(args.verbose)

        # check for state file extension
        if args.state_file is None:
            raise ValueError(
                "Specify project file name (with .asreview extension).")

        # print intro message
        print(ASCII_LOGO + ASCII_MSG_SIMULATE)

        # for webapp
        if args.dataset == "":

            project = ASReviewProject(args.state_file)

            with open_state(args.state_file) as state:
                settings = state.settings

                # Check if there are new labeled records.
                exist_new_labeled_records = state.exist_new_labeled_records

            # collect command line arguments and pass them to the reviewer
            if exist_new_labeled_records:
                as_data = read_data(project)
                prior_idx = args.prior_idx

            classifier_model = get_classifier(settings.model)
            query_model = get_query_model(settings.query_strategy)
            balance_model = get_balance_model(settings.balance_strategy)
            feature_model = get_feature_model(settings.feature_extraction)

        # for simulation CLI
        else:

            # do this check now and again when zipping.
            if Path(args.state_file).exists():
                raise ProjectExistsError("Project already exists.")

            as_data = load_data(args.dataset)

            if len(as_data) == 0:
                raise ValueError("Supply at least one dataset"
                                 " with at least one record.")

            # create a project file
            fp_tmp_simulation = Path(
                args.state_file).with_suffix(".asreview.tmp")

            project = ASReviewProject.create(
                fp_tmp_simulation,
                project_id=Path(args.state_file).stem,
                project_mode="simulate",
                project_name=Path(args.state_file).stem,
                project_description="Simulation created via ASReview via "
                                    "command line interface"
            )

            # Add the dataset to the project file.
            dataset_path = _get_dataset_path_from_args(args.dataset)

            as_data.to_file(
                Path(fp_tmp_simulation, 'data', dataset_path)
            )
            # Update the project.json.
            project.update_config(dataset_path=dataset_path)

            # create a new settings object from arguments
            settings = ASReviewSettings(
                model=args.model,
                n_instances=args.n_instances,
                stop_if=args.stop_if,
                n_prior_included=args.n_prior_included,
                n_prior_excluded=args.n_prior_excluded,
                query_strategy=args.query_strategy,
                balance_strategy=args.balance_strategy,
                feature_extraction=args.feature_extraction,
                mode="simulate",
                data_fp=None)
            settings.from_file(args.config_file)

            # Initialize models.
            random_state = get_random_state(args.seed)
            classifier_model = get_classifier(settings.model,
                                              random_state=random_state,
                                              **settings.model_param)
            query_model = get_query_model(settings.query_strategy,
                                          random_state=random_state,
                                          **settings.query_param)
            balance_model = get_balance_model(settings.balance_strategy,
                                              random_state=random_state,
                                              **settings.balance_param)
            feature_model = get_feature_model(settings.feature_extraction,
                                              random_state=random_state,
                                              **settings.feature_param)

            # TODO{Deprecate and integrate with the model}
            # LSTM models need embedding matrices.
            if classifier_model.name.startswith("lstm-"):
                texts = as_data.texts
                classifier_model.embedding_matrix = feature_model.\
                    get_embedding_matrix(texts, args.embedding_fp)

            # prior knowledge
            if args.prior_idx is not None and args.prior_record_id is not None and \
                    len(args.prior_idx) > 0 and len(args.prior_record_id) > 0:
                raise ValueError(
                    "Not possible to provide both prior_idx and prior_record_id"
                )

            prior_idx = args.prior_idx
            if args.prior_record_id is not None and len(
                    args.prior_record_id) > 0:
                prior_idx = convert_id_to_idx(as_data, args.prior_record_id)

            print("The following records are prior knowledge:\n")
            for prior_record_id in args.prior_record_id:
                preview = preview_record(as_data.record(prior_record_id))
                print(f"{prior_record_id} - {preview}")

        try:
            # Initialize the review class.
            reviewer = ReviewSimulate(
                as_data,
                project=project,
                model=classifier_model,
                query_model=query_model,
                balance_model=balance_model,
                feature_model=feature_model,
                n_papers=args.n_papers,
                n_instances=args.n_instances,
                stop_if=args.stop_if,
                prior_indices=prior_idx,
                n_prior_included=args.n_prior_included,
                n_prior_excluded=args.n_prior_excluded,
                init_seed=args.init_seed,
                write_interval=args.write_interval)

            # Start the review process.
            project.update_review(status="review")

            reviewer.review()
        except Exception as err:

            # save the error to the project
            project.set_error(err)

            raise err

        print("Simulation finished.")
        project.mark_review_finished()

        # create .ASReview file out of simulation folder
        if args.dataset != "":

            project.export(args.state_file)
            shutil.rmtree(fp_tmp_simulation)


DESCRIPTION_SIMULATE = """
ASReview for simulation.

The simulation modus is used to measure the performance of the ASReview
software on existing systematic reviews. The software shows how many
papers you could have potentially skipped during the systematic
review."""


def _simulate_parser(prog="simulate", description=DESCRIPTION_SIMULATE):
    parser = _base_parser(prog=prog, description=description)
    # Active learning parameters
    # File path to the data.
    parser.add_argument(
        "dataset",
        type=str,
        help="File path to the dataset or one of the benchmark datasets.")
    # Initial data (prior knowledge)
    parser.add_argument("--n_prior_included",
                        default=DEFAULT_N_PRIOR_INCLUDED,
                        type=int,
                        help="Sample n prior included papers. "
                        "Only used when --prior_idx is not given. "
                        f"Default {DEFAULT_N_PRIOR_INCLUDED}")

    parser.add_argument("--n_prior_excluded",
                        default=DEFAULT_N_PRIOR_EXCLUDED,
                        type=int,
                        help="Sample n prior excluded papers. "
                        "Only used when --prior_idx is not given. "
                        f"Default {DEFAULT_N_PRIOR_EXCLUDED}")

    parser.add_argument(
        "--prior_idx",
        default=[],
        nargs="*",
        type=int,
        help="Prior indices by rownumber (0 is first rownumber).")
    parser.add_argument("--prior_record_id",
                        default=[],
                        nargs="*",
                        type=int,
                        help="Prior indices by record_id.")
    # logging and verbosity
    parser.add_argument(
        "--state_file",
        "-s",
        default=None,
        type=str,
        help="Location to ASReview project file of simulation.")
    parser.add_argument("-m",
                        "--model",
                        type=str,
                        default=DEFAULT_MODEL,
                        help=f"The prediction model for Active Learning. "
                        f"Default: '{DEFAULT_MODEL}'.")
    parser.add_argument("-q",
                        "--query_strategy",
                        type=str,
                        default=DEFAULT_QUERY_STRATEGY,
                        help=f"The query strategy for Active Learning. "
                        f"Default: '{DEFAULT_QUERY_STRATEGY}'.")
    parser.add_argument(
        "-b",
        "--balance_strategy",
        type=str,
        default=DEFAULT_BALANCE_STRATEGY,
        help="Data rebalancing strategy mainly for RNN methods. Helps against"
        " imbalanced dataset with few inclusions and many exclusions. "
        f"Default: '{DEFAULT_BALANCE_STRATEGY}'")
    parser.add_argument(
        "-e",
        "--feature_extraction",
        type=str,
        default=DEFAULT_FEATURE_EXTRACTION,
        help="Feature extraction method. Some combinations of feature"
        " extraction method and prediction model are impossible/ill"
        " advised."
        f"Default: '{DEFAULT_FEATURE_EXTRACTION}'")
    parser.add_argument(
        '--init_seed',
        default=None,
        type=int,
        help="Seed for setting the prior indices if the --prior_idx option is "
        "not used. If the option --prior_idx is used with one or more "
        "index, this option is ignored.")
    parser.add_argument("--n_instances",
                        default=DEFAULT_N_INSTANCES,
                        type=int,
                        help="Number of papers queried each query."
                        f"Default {DEFAULT_N_INSTANCES}.")
    parser.add_argument(
        "--n_queries",
        type=type_n_queries,
        default="min",
        help="Deprecated, use 'stop_if' instead.")
    parser.add_argument(
        "--stop_if",
        type=type_n_queries,
        default="min",
        help="The number of label actions to simulate. Default, 'min' "
        "will stop simulating when all relevant records are found. Use -1 "
        "to simulate all labels actions.")
    parser.add_argument(
        "-n",
        "--n_papers",
        type=int,
        default=None,
        help="Deprecated, use 'stop_if' instead.")
    parser.add_argument("--verbose",
                        "-v",
                        default=0,
                        type=int,
                        help="Verbosity")
    parser.add_argument(
        "--write_interval",
        "-w",
        default=None,
        type=int,
        help="The simulation data will be written away after each set of this"
        "many labeled records. By default only writes away data at the end"
        "of the simulation to make it as fast as possible.")

    return parser



# Copyright 2019-2022 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from math import floor
from math import log

import numpy as np

from asreview.models.balance.base import BaseBalance
from asreview.models.balance.simple import SimpleBalance
from asreview.utils import get_random_state


class DoubleBalance(BaseBalance):
    """Double balance strategy.

    Class to get the two way rebalancing function and arguments.
    It super samples ones depending on the number of 0's and total number
    of samples in the training data.

    Arguments
    ---------
    a: float
        Governs the weight of the 1's. Higher values mean linearly more 1's
        in your training sample.
    alpha: float
        Governs the scaling the weight of the 1's, as a function of the
        ratio of ones to zeros. A positive value means that the lower the
        ratio of zeros to ones, the higher the weight of the ones.
    b: float
        Governs how strongly we want to sample depending on the total
        number of samples. A value of 1 means no dependence on the total
        number of samples, while lower values mean increasingly stronger
        dependence on the number of samples.
    beta: float
        Governs the scaling of the weight of the zeros depending on the
        number of samples. Higher values means that larger samples are more
        strongly penalizing zeros.
    """

    name = "double"
    label = "Dynamic resampling (Double)"

    def __init__(self,
                 a=2.155,
                 alpha=0.94,
                 b=0.789,
                 beta=1.0,
                 random_state=None):
        super(DoubleBalance, self).__init__()
        self.a = a
        self.alpha = alpha
        self.b = b
        self.beta = beta
        self.fallback_model = SimpleBalance()
        self._random_state = get_random_state(random_state)

    def sample(self, X, y, train_idx):
        """Resample the training data.

        Arguments
        ---------
        X: numpy.ndarray
            Complete feature matrix.
        y: numpy.ndarray
            Labels for all papers.
        train_idx: numpy.ndarray
            Training indices, that is all papers that have been reviewed.

        Returns
        -------
        numpy.ndarray,numpy.ndarray:
            X_train, y_train: the resampled matrix, labels.
        """
        # Get inclusions and exclusions
        one_idx = train_idx[np.where(y[train_idx] == 1)]
        zero_idx = train_idx[np.where(y[train_idx] == 0)]

        # Fall back to simple sampling if we have only ones or zeroes.
        if len(one_idx) == 0 or len(zero_idx) == 0:
            self.fallback_model.sample(X, y, train_idx)

        n_one = len(one_idx)
        n_zero = len(zero_idx)
        n_train = n_one + n_zero

        # Compute sampling weights.
        one_weight = _one_weight(n_one, n_zero, self.a, self.alpha)
        zero_weight = _zero_weight(n_one + n_zero, self.b, self.beta)
        tot_zo_weight = one_weight * n_one + zero_weight * n_zero
        # Number of inclusions to sample.
        n_one_train = random_round(
            one_weight * n_one * n_train / tot_zo_weight, self._random_state)
        # Should be at least 1, and at least two spots should be for exclusions.
        n_one_train = max(1, min(n_train - 2, n_one_train))
        # Number of exclusions to sample
        n_zero_train = n_train - n_one_train

        # Sample records of ones and zeroes
        one_train_idx = fill_training(one_idx, n_one_train, self._random_state)
        zero_train_idx = fill_training(zero_idx, n_zero_train,
                                       self._random_state)
        # Merge and shuffle.
        all_idx = np.concatenate([one_train_idx, zero_train_idx])
        self._random_state.shuffle(all_idx)

        # Return resampled feature matrix and labels.
        return X[all_idx], y[all_idx]

    def full_hyper_space(self):
        from hyperopt import hp
        parameter_space = {
            "bal_a": hp.lognormal("bal_a", 0, 1),
            "bal_alpha": hp.uniform("bal_alpha", 0, 2),
            "bal_b": hp.uniform("bal_b", 0, 1),
            # "bal_beta": hp.uniform("bal_beta", 0, 2),
        }
        return parameter_space, {}


def _one_weight(n_one, n_zero, a, alpha):
    """Get the weight of the ones."""
    weight = a * (n_one / n_zero)**(-alpha)
    return weight


def _zero_weight(n_read, b, beta):
    """Get the weight of the zeros."""
    weight = 1 - (1 - b) * (1 + log(n_read))**(-beta)
    return weight


def random_round(value, random_state):
    """Round up or down, depending on how far the value is.

    For example: 8.1 would be rounded to 8, 90% of the time, and rounded
    to 9, 10% of the time.
    """
    base = int(floor(value))
    if random_state.rand() < value - base:
        base += 1
    return base


def fill_training(src_idx, n_train, random_state):
    """Copy/sample until there are n_train indices sampled/copied.
    """
    # Number of copies needed.
    n_copy = int(n_train / len(src_idx))
    # For the remainder, use sampling.
    n_sample = n_train - n_copy * len(src_idx)

    # Copy indices
    dest_idx = np.tile(src_idx, n_copy).reshape(-1)
    # Add samples
    dest_idx = np.append(dest_idx,
                         random_state.choice(src_idx, n_sample, replace=False))
    return dest_idx



# Copyright 2019-2022 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import abstractmethod

import numpy as np
from scipy.sparse import hstack
from scipy.sparse import issparse

from asreview.models.base import BaseModel


class BaseFeatureExtraction(BaseModel):
    """Base class for feature extraction methods."""
    name = "base-feature"

    def __init__(self, split_ta=0, use_keywords=0):
        self.split_ta = split_ta
        self.use_keywords = use_keywords

    def fit_transform(self, texts, titles=None, abstracts=None, keywords=None):
        """Fit and transform a list of texts.

        Arguments
        ---------
        texts: numpy.ndarray
            A sequence of texts to be transformed. They are not yet tokenized.

        Returns
        -------
        numpy.ndarray
            Feature matrix representing the texts.
        """
        self.fit(texts)
        if self.split_ta > 0:
            if titles is None or abstracts is None:
                raise ValueError("Error: if splitting titles and abstracts,"
                                 " supply them!")
            X_titles = self.transform(titles)
            X_abstracts = self.transform(abstracts)
            if issparse(X_titles) and issparse(X_abstracts):
                X = hstack([X_titles, X_abstracts]).tocsr()
            else:
                X = np.concatenate((X_titles, X_abstracts), axis=1)
        else:
            X = self.transform(texts)

        if self.use_keywords and keywords is not None:
            join_keys = np.array([" ".join(key) for key in keywords])
            X_keywords = self.transform(join_keys)
            if issparse(X_keywords):
                X = hstack([X, X_keywords]).tocsr()
            else:
                X = np.concatenate((X, X_keywords), axis=1)

        return X

    def fit(self, texts):
        """Fit the model to the texts.

        It is not always necessary to implement this if there's not real
        fitting being done.

        Arguments
        ---------
        texts: numpy.ndarray
            Texts to be fitted.
        """
        pass

    @abstractmethod
    def transform(self, texts):
        """Transform a list of texts.

        Arguments
        ---------
        texts: numpy.ndarray
            A sequence of texts to be transformed. They are not yet tokenized.

        Returns
        -------
        numpy.ndarray
            Feature matrix representing the texts.
        """
        raise NotImplementedError

    def full_hyper_space(self):
        from hyperopt import hp
        hyper_choices = {}
        hyper_space = {
            "fex_split_ta": hp.randint("fex_split_ta", 2),
            "fex_use_keywords": hp.randint("fex_use_keywords", 2),
        }
        return hyper_space, hyper_choices



# Copyright 2019-2022 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from sklearn.naive_bayes import MultinomialNB

from asreview.models.classifiers.base import BaseTrainClassifier


class NaiveBayesClassifier(BaseTrainClassifier):
    """Naive Bayes classifier.

    Naive Bayes classifier. Only works in combination with the
    :class:`asreview.models.feature_extraction.Tfidf` feature extraction model.
    Though relatively simplistic, seems to work quite well on a wide range of
    datasets.

    The naive Bayes classifier is an implementation based
    on the sklearn multinomial naive Bayes classifier.

    Arguments
    ---------
    alpha : float, default=3.822
        Additive (Laplace/Lidstone) smoothing parameter
        (0 for no smoothing).
    """

    name = "nb"
    label = "Naive Bayes"

    def __init__(self, alpha=3.822):

        super(NaiveBayesClassifier, self).__init__()
        self.alpha = alpha
        self._model = MultinomialNB(alpha=alpha)
        logging.debug(self._model)

    def full_hyper_space(self):
        from hyperopt import hp
        hyper_choices = {}
        hyper_space = {
            "mdl_alpha": hp.lognormal("mdl_alpha", 0, 1),
        }
        return hyper_space, hyper_choices



# Copyright 2019-2022 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC
from abc import abstractmethod


class BaseState(ABC):
    def __init__(self, read_only=False):
        """Abstract Base Class for state.

        read_only: bool
            Whether to open file in read only mode.
        """
        self.read_only = read_only

    def __enter__(self):
        return self

    def __exit__(self, *_, **__):
        self.close()

    def __str__(self):
        return str(self.to_dict())

    @abstractmethod
    def _create_new_state_file(self, working_dir, review_id):
        """Create empty internal structure for state.

        Arguments
        ---------
        working_dir: str, pathlib.Path
            Location of project file.
        review_id: str
            Identifier of the review.
        """
        raise NotImplementedError

    @abstractmethod
    def _restore(self, working_dir, review_id):
        """Restore state from a state file.

        Arguments
        ---------
        working_dir: str, pathlib.Path
            Location of project file.
        review_id: str
            Identifier of the review.
        """
        raise NotImplementedError

    @abstractmethod
    def add_record_table(self, record_ids):
        """Add the record table to the state.

        Arguments
        ---------
        record_ids: list-like
            List containing all record ids of the dataset.
        """
        raise NotImplementedError

    @abstractmethod
    def add_last_probabilities(self, probabilities):
        """Save the probabilities produced by the last classifier.

        Arguments
        ---------
        probabilities: list-like.
            List containing the probabilities for every record.
        """
        raise NotImplementedError

    @abstractmethod
    def add_last_ranking(self, ranked_record_ids, classifier,
                         query_strategy, balance_strategy, feature_extraction,
                         training_set):
        """Save the ranking of the last iteration of the model, in the ranking
        order.

        Arguments
        ---------
        ranked_record_ids: list, numpy.ndarray
            A list of records ids in the order that they were ranked.
        classifier: str
            Name of the classifier of the model.
        query_strategy: str
            Name of the query strategy of the model.
        balance_strategy: str
            Name of the balance strategy of the model.
        feature_extraction: str
            Name of the feature extraction method of the model.
        training_set: int
            Number of labeled records available at the time of training.
        """
        raise NotImplementedError

    @abstractmethod
    def get_last_probabilities(self):
        """Get the probabilities produced by the last classifier.

        Returns
        -------
        pd.Series:
            Series with name 'proba' containing the probabilities.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def settings(self):
        """Get settings from the state.
        """
        raise NotImplementedError

    @abstractmethod
    def add_note(self, note, record_id):
        """Add a text note to save with a labeled record.

        Arguments
        ---------
        note: str
            Text note to save.
        record_id: int
            Identifier of the record to which the note should be added.
        """
        raise NotImplementedError

    @abstractmethod
    def add_labeling_data(self, record_ids, labels, notes=None, prior=False):
        """Add the data corresponding to a labeling action to the state file.

        Arguments
        ---------
        record_ids: list, numpy.ndarray
            A list of ids of the labeled records as int.
        labels: list, numpy.ndarray
            A list of labels of the labeled records as int.
        notes: list of str/None
            A list of text notes to save with the labeled records.
        prior: bool
            Whether the added record are prior knowledge.
        """
        raise NotImplementedError

    def update_decision(self, record_id, label, note=None):
        """Change the label of an already labeled record.

        Arguments
        ---------
        record_id: int
            Id of the record whose label should be changed.
        label: 0 / 1
            New label of the record.
        note: str
            Note to add to the record.
        """
        raise NotImplementedError

    def get_decision_changes(self):
        """Get the record ids of the records whose labels have been changed
        after the original labeling action."""
        raise NotImplementedError

    @abstractmethod
    def get_pool_labeled_pending(self):
        """Return the labeled and unlabeled records and the records pending a
        labeling decision.

        Returns
        -------
        tuple (pd.Series, pd.DataFrame, pd.Series):
            Returns a tuple (pool, labeled, pending). Pool is a series
            containing the unlabeled, not pending record_ids, ordered by the
            last predicted ranking of the model. Labeled is a dataframe
            containing the record_ids and labels of the labeled records, in the
            order that they were labeled. Pending is a series containing the
            record_ids of the records whose label is pending.
        """
        raise NotImplementedError

    def is_empty(self):
        """Check if state has no results.

        Returns
        -------
        bool
            True if empty.
        """
        return self.n_records_labeled == 0

    @property
    @abstractmethod
    def n_records_labeled(self):
        """Number labeled records.

        Returns
        -------
        int
            Number of labeled records, priors counted individually.
        """
        raise NotImplementedError

    @abstractmethod
    def get_dataset(self, columns=None):
        """Get a column from the results table.

        Arguments
        ---------
        columns: list
            List of columns names of the results table.

        Returns
        -------
        pd.DataFrame:
            Dataframe containing the data of the specified columns of the
            results table.
        """
        raise NotImplementedError

    def get_order_of_labeling(self):
        """Get full array of record id's in order that they were labeled.

        Returns
        -------
        pd.Series:
            The record_id's in the order that they were labeled.
        """
        raise NotImplementedError

    def get_labels(self):
        """Get the labels from the state file.

        Returns
        -------
        pd.Series:
            Series containing the labels at each labelling moment.
        """
        raise NotImplementedError

    def get_classifiers(self):
        """Get the classifiers from the state file.

        Returns
        -------
        pd.Series:
            Series containing the classifier used at each labeling moment.
        """
        raise NotImplementedError

    def get_query_strategies(self):
        """Get the query strategies from the state file.

        Returns
        -------
        pd.Series:
            Series containing the query strategy used to get the record to
            query at each labeling moment.
        """
        raise NotImplementedError

    def get_balance_strategies(self):
        """Get the balance strategies from the state file.

        Returns
        -------
        pd.Series:
            Series containing the balance strategy used to get the training
            data at each labeling moment.
        """
        raise NotImplementedError

    def get_feature_extraction(self):
        """Get the query strategies from the state file.

        Returns
        -------
        pd.Series:
            Series containing the feature extraction method used for the
            classifier input at each labeling moment.
        """
        raise NotImplementedError

    def get_training_sets(self):
        """Get the training_sets from the state file.

        Returns
        -------
        pd.Series:
            Series containing the training set on which the classifier was
            fit at each labeling moment.
        """
        raise NotImplementedError

    def get_labeling_times(self, time_format='int'):
        """Get the time of labeling the state file.

        Arguments
        ---------
        time_format: 'int' or 'datetime'
            Format of the return value. If it is 'int' you get a UTC timestamp,
            if it is 'datetime' you get datetime instead of an integer.

        Returns
        -------
        pd.Series:
            If format='int' you get a UTC timestamp (integer number of
            microseconds), if it is 'datetime' you get datetime format.
        """
        raise NotImplementedError

    @abstractmethod
    def close(self):
        """Close the files opened by the state.

        """
        raise NotImplementedError

    def to_dict(self):
        """Convert state to dictionary.

        Returns
        -------
        dict:
            Dictionary with all settings and results.
        """
        state_data = self.get_dataset()
        state_dict = {
            'settings': vars(self.settings),
            'data': state_data.to_dict()
        }
        return state_dict



# Copyright 2019-2022 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from io import BytesIO
from urllib.request import urlopen

import pytest

from asreview.webapp.tests.utils import retrieve_project_url_github

# Retrieve urls to .asreview files exported from previous versions
project_urls = retrieve_project_url_github()


@pytest.mark.parametrize("url", project_urls)
def test_project_file(tmp_path, client, url):
    """Test import and continue a project created in previous versions."""

    # change default folder for projects
    os.environ["ASREVIEW_PATH"] = str(tmp_path)

    # Test import uploaded project
    with urlopen(url) as project_file:
        response_import = client.post(
            "/api/projects/import_project",
            data={"file": (BytesIO(project_file.read()), "project.asreview")})
    json_data_import = response_import.get_json()
    assert response_import.status_code == 200

    project_id = json_data_import["id"]
    api_url = f"/api/projects/{project_id}"

    # Test get dashboard analytics
    response_stats = client.get("/api/projects/stats")
    json_data_stats = response_stats.get_json()

    assert "result" in json_data_stats
    assert "n_in_review" in json_data_stats["result"]
    assert "n_finished" in json_data_stats["result"]
    assert "n_reviewed" in json_data_stats["result"]
    assert "n_included" in json_data_stats["result"]
    assert isinstance(json_data_stats["result"], dict)

    # Test get projects
    response_projects = client.get("/api/projects")
    json_data_projects = response_projects.get_json()
    assert "result" in json_data_projects
    assert any(item["id"] == project_id
               for item in json_data_projects["result"])

    # Test export project before upgrade
    response_export_old_project = client.get(f"{api_url}/export_project")
    assert response_export_old_project.status_code == 200

    # Test upgrade project if old
    response_upgrade_if_old = client.get(f"{api_url}/upgrade_if_old")
    assert response_upgrade_if_old.status_code == 200

    # Test get info on the project
    response_get_info = client.get(f"{api_url}/info")
    json_data_get_info = response_get_info.get_json()
    assert json_data_get_info["id"] == project_id

    # Test get dataset writer
    response_get_writer = client.get(f"{api_url}/dataset_writer")
    json_data_get_writer = response_get_writer.get_json()
    assert isinstance(json_data_get_writer["result"], list)

    # Test update info of the project
    response_update_info = client.put(f"{api_url}/info",
                                      data={
                                          "mode": "explore",
                                          "name": json_data_get_info["name"],
                                          "authors":
                                          json_data_get_info["authors"],
                                          "description": "Hoi Elas"
                                      })
    assert response_update_info.status_code == 200

    # Test get progress info on the article
    response_progress = client.get(f"{api_url}/progress")
    json_data_progress = response_progress.get_json()
    assert isinstance(json_data_progress, dict)

    # Test get progress density of the project
    response_progress_density = client.get(f"{api_url}/progress_density")
    json_data_progress_density = response_progress_density.get_json()
    assert "relevant" in json_data_progress_density
    assert "irrelevant" in json_data_progress_density
    assert isinstance(json_data_progress_density, dict)

    # Test get cumulative number of inclusions by ASReview/at random
    response_progress_recall = client.get(f"{api_url}/progress_recall")
    json_data_progress_recall = response_progress_recall.get_json()
    assert "asreview" in json_data_progress_recall
    assert "random" in json_data_progress_recall
    assert isinstance(json_data_progress_recall, dict)

    # Test retrieve documents in order of review
    response_get_document = client.get(f"{api_url}/get_document")
    json_data_get_document = response_get_document.get_json()
    assert "result" in json_data_get_document
    assert isinstance(json_data_get_document, dict)

    # get doc_id from the queue and label the item
    doc_id = json_data_get_document["result"]["doc_id"]

    # Test retrieve classification result
    response_classify_instance = client.post(f"{api_url}/record/{doc_id}",
                                             data={
                                                 "doc_id": doc_id,
                                                 "label": 1,
                                             })
    assert response_classify_instance.status_code == 200

    # Test update classification result
    response_update_classify = client.put(f"{api_url}/record/{doc_id}",
                                          data={
                                              "doc_id": doc_id,
                                              "label": 0,
                                          })
    assert response_update_classify.status_code == 200

    # Test retrieve review history
    response_prior = client.get(f"{api_url}/labeled")
    json_data_prior = response_prior.get_json()
    assert "result" in json_data_prior
    assert isinstance(json_data_prior["result"], list)

    # Test export result
    response_export_result_csv = client.get(
        f"{api_url}/export_dataset?file_format=csv")
    response_export_result_tsv = client.get(
        f"{api_url}/export_dataset?file_format=tsv")
    response_export_result_excel = client.get(
        f"{api_url}/export_dataset?file_format=xlsx")
    assert response_export_result_csv.status_code == 200
    assert response_export_result_tsv.status_code == 200
    assert response_export_result_excel.status_code == 200

    # Test export project
    response_export_project = client.get(f"{api_url}/export_project")
    assert response_export_project.status_code == 200

    # Test get project status
    response_status = client.get(f"{api_url}/status")
    json_data_status = response_status.get_json()
    assert "status" in json_data_status
    assert isinstance(json_data_status, dict)

    # Test mark project as finished
    response_finish = client.put(f"{api_url}/status",
                                 data={"status": "finished"})
    assert response_finish.status_code == 200

    # Test delete project
    response_delete = client.delete(f"{api_url}/delete")
    assert response_delete.status_code == 200



# Copyright 2019-2022 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from collections import OrderedDict
from datetime import datetime
from pathlib import Path

from asreview.settings import ASReviewSettings
from asreview.state.legacy.dict import DictState


class JSONState(DictState):
    """Class for storing the state of a Systematic Review using JSON files."""
    version = "2.1"

    def __init__(self, state_fp, read_only=False):
        super(JSONState, self).__init__(state_fp)
        self.read_only = read_only

    def save(self):
        if self.read_only:
            raise ValueError("State error: trying to save when opened file"
                             " in read_only mode.")
        self._state_dict["time"]["end_time"] = str(datetime.now())
        fp = Path(self.state_fp)

        if fp.is_file:
            fp.parent.mkdir(parents=True, exist_ok=True)

        try:
            self._state_dict["current_queries"] = {
                str(key): val
                for key, val in self._state_dict["current_queries"].items()
            }
        except KeyError:
            pass

        with fp.open('w') as outfile:
            json.dump(self._state_dict, outfile, indent=2)

        try:
            self._state_dict["current_queries"] = {
                int(key): val
                for key, val in self._state_dict["current_queries"].items()
            }
        except KeyError:
            pass

    def restore(self, fp):
        try:
            with open(fp, "r") as f:
                self._state_dict = OrderedDict(json.load(f))
            state_version = self._state_dict["version"]
            if state_version != self.version:
                raise ValueError(
                    f"State cannot be read: state version {self.version}, "
                    f"state file version {state_version}.")
            self.settings = ASReviewSettings(**self._state_dict["settings"])
            try:
                self._state_dict["current_queries"] = {
                    int(key): val
                    for key, val in
                    self._state_dict["current_queries"].items()
                }
            except KeyError:
                pass
        except FileNotFoundError:
            self.initialize_structure()



# Copyright 2019-2022 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from sklearn.svm import SVC

from asreview.models.classifiers.base import BaseTrainClassifier
from asreview.models.classifiers.utils import _set_class_weight


class SVMClassifier(BaseTrainClassifier):
    """
    Support vector machine classifier.

    The Support Vector Machine classifier is an implementation based
    on the sklearn Support Vector Machine classifier.

    Arguments
    ---------
    gamma: str
        Gamma parameter of the SVM model.
    class_weight: float
        class_weight of the inclusions.
    C: float
        C parameter of the SVM model.
    kernel: str
        SVM kernel type.
    random_state: int, RandomState
        State of the RNG.
    """

    name = "svm"
    label = "Support vector machine"

    def __init__(self,
                 gamma="auto",
                 class_weight=0.249,
                 C=15.4,
                 kernel="linear",
                 random_state=None):
        super(SVMClassifier, self).__init__()
        self.gamma = gamma
        self.class_weight = class_weight
        self.C = C
        self.kernel = kernel
        self._random_state = random_state

        self._model = SVC(
            kernel=kernel,
            C=C,
            class_weight=_set_class_weight(class_weight),
            random_state=random_state,
            gamma=gamma,
            probability=True)

    def full_hyper_space(self):
        from hyperopt import hp
        hyper_choices = {
            "mdl_gamma": ["auto", "scale"],
            "mdl_kernel": ["linear", "rbf", "poly", "sigmoid"]
        }

        hyper_space = {
            "mdl_gamma": hp.choice('mdl_gamma', hyper_choices["mdl_gamma"]),
            "mdl_kernel": hp.choice('mdl_kernel', hyper_choices["mdl_kernel"]),
            "mdl_C": hp.lognormal('mdl_C', 0, 2),
            "mdl_class_weight": hp.lognormal('mdl_class_weight', 0, 1)
        }
        return hyper_space, hyper_choices



# Copyright 2019-2022 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from urllib.request import urlopen


def retrieve_project_url_github(major=None):
    '''Retrieve .asreview file url from
    asreview-project-files-testing GitHub repository'''

    repo = "/asreview/asreview-project-files-testing"
    repo_api_url = "https://api.github.com/repos" + repo + "/git/trees/master"
    repo_url = "https://github.com" + repo + "/blob/master"
    file_type = "startreview.asreview?raw=true"

    json_file = json.loads(urlopen(repo_api_url).read().decode("utf-8"))["tree"]

    version_tags = []
    project_urls = []

    for file in json_file:
        if file["type"] == "tree":
            version_tags.append(file["path"])

    for tag in version_tags:
        file_version = f"/{tag}/asreview-project-{tag.replace('.', '-')}-"

        if major is None or int(tag[1]) == major:
            project_urls.append(repo_url + file_version + file_type)

    return project_urls



# Copyright 2019-2022 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pandas as pd

from asreview.io.utils import _standardize_dataframe


class CSVReader():
    """CVS file reader.
    """

    read_format = [".csv", ".tab", ".tsv"]
    write_format = [".csv", ".tsv", ".xlsx"]

    @classmethod
    def read_data(cls, fp):
        """Import dataset.

        Arguments
        ---------
        fp: str, pathlib.Path
            File path to the CSV file.

        Returns
        -------
        list:
            List with entries.
        """
        for encoding in ["utf-8", "ISO-8859-1"]:
            try:
                df = pd.read_csv(
                    fp,
                    sep=None,
                    encoding=encoding,
                    engine='python'
                )
                return _standardize_dataframe(df)
            except UnicodeDecodeError:
                # if unicode error, go to next encoding
                continue

        raise UnicodeDecodeError("The encoding of the file is not supported.")



# Copyright 2019-2022 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


def test_landing(client):
    """Test if index.html is available.

    This test will fail if build is missing. Please run
    `python setup.py compile_assets` first.
    """
    response = client.get("/")
    html = response.data.decode()

    assert "<title>ASReview LAB - A tool for AI-assisted systematic reviews</title>" in html  # noqa


def test_boot(client):
    """Test if version number is available on boot."""
    response = client.get("/boot")
    json_data = response.get_json()

    assert "version" in json_data
    assert "status" in json_data



# Copyright 2019-2022 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

try:
    from gensim.models.doc2vec import Doc2Vec as GenSimDoc2Vec
    from gensim.models.doc2vec import TaggedDocument
    from gensim.utils import simple_preprocess
    GENSIM_AVAILABLE = True
except ImportError:
    GENSIM_AVAILABLE = False

from asreview.models.feature_extraction.base import BaseFeatureExtraction


def _check_gensim():
    if not GENSIM_AVAILABLE:
        raise ImportError(
            "Install gensim package to use"
            " Doc2Vec.")


def _train_model(corpus, *args, **kwargs):

    model = GenSimDoc2Vec(*args, **kwargs)
    model.build_vocab(corpus)
    model.train(corpus, total_examples=model.corpus_count, epochs=model.epochs)
    return model


def _transform_text(model, corpus):
    X = []
    for doc_id in range(len(corpus)):
        doc_vec = model.infer_vector(corpus[doc_id].words)
        X.append(doc_vec)
    return np.array(X)


class Doc2Vec(BaseFeatureExtraction):
    """Doc2Vec feature extraction technique.

    Feature extraction technique provided by the `gensim
    <https://radimrehurek.com/gensim/>`__ package. It takes relatively long to
    create a feature matrix with this method. However, this only has to be
    done once per simulation/review. The upside of this method is the
    dimension- reduction that generally takes place, which makes the modelling
    quicker.

    .. note::

        This feature extraction technique requires ``gensim`` to be installed.
        Use ``pip install gensim`` or install all optional ASReview
        dependencies with ``pip install asreview[all]``

    Arguments
    ---------
    vector_size: int
        Output size of the vector.
    epochs: int
        Number of epochs to train the doc2vec model.
    min_count: int
        Minimum number of occurences for a word in the corpus for it to
        be included in the model.
    n_jobs: int
        Number of threads to train the model with.
    window: int
        Maximum distance over which word vectors influence each other.
    dm_concat: int
        Whether to concatenate word vectors or not.
        See paper for more detail.
    dm: int
        Model to use.
        0: Use distribute bag of words (DBOW).
        1: Use distributed memory (DM).
        2: Use both of the above with half the vector size and concatenate
        them.
    dbow_words: int
        Whether to train the word vectors using the skipgram method.
    """

    name = "doc2vec"
    label = "Doc2Vec"

    def __init__(self,
                 *args,
                 vector_size=40,
                 epochs=33,
                 min_count=1,
                 n_jobs=1,
                 window=7,
                 dm_concat=0,
                 dm=2,
                 dbow_words=0,
                 **kwargs):
        """Initialize the doc2vec model."""
        super(Doc2Vec, self).__init__(*args, **kwargs)
        self.vector_size = int(vector_size)
        self.epochs = int(epochs)
        self.min_count = int(min_count)
        self.n_jobs = int(n_jobs)
        self.window = int(window)
        self.dm_concat = int(dm_concat)
        self.dm = int(dm)
        self.dbow_words = int(dbow_words)
        self._model = None
        self._model_dm = None
        self._model_dbow = None

    def fit(self, texts):

        # check is gensim is available
        _check_gensim()

        model_param = {
            "vector_size": self.vector_size,
            "epochs": self.epochs,
            "min_count": self.min_count,
            "workers": self.n_jobs,
            "window": self.window,
            "dm_concat": self.dm_concat,
            "dbow_words": self.dbow_words,
        }

        corpus = [
            TaggedDocument(simple_preprocess(text), [i])
            for i, text in enumerate(texts)
        ]

        # If self.dm is 2, train both models and concatenate the feature
        # vectors later. Resulting vector size should be the same.
        if self.dm == 2:
            model_param["vector_size"] = int(model_param["vector_size"] / 2)
            self.model_dm = _train_model(corpus, **model_param, dm=1)
            self.model_dbow = _train_model(corpus, **model_param, dm=0)
        else:
            self.model = _train_model(corpus, **model_param, dm=self.dm)

    def transform(self, texts):

        # check is gensim is available
        _check_gensim()

        corpus = [
            TaggedDocument(simple_preprocess(text), [i])
            for i, text in enumerate(texts)
        ]

        if self.dm == 2:
            X_dm = _transform_text(self.model_dm, corpus)
            X_dbow = _transform_text(self.model_dbow, corpus)
            X = np.concatenate((X_dm, X_dbow), axis=1)
        else:
            X = _transform_text(self.model, corpus)
        return X

    def full_hyper_space(self):
        from hyperopt import hp
        eps = 1e-7

        hyper_space, hyper_choices = super(Doc2Vec, self).full_hyper_space()
        hyper_space.update({
            "fex_vector_size":
            hp.quniform("fex_vector_size", 31.5, 127.5 - eps, 8),
            "fex_epochs":
            hp.quniform("fex_epochs", 20, 50, 1),
            "fex_min_count":
            hp.quniform("fex_min_count", 0.5, 2.499999, 1),
            "fex_window":
            hp.quniform("fex_window", 4.5, 9.4999999, 1),
            "fex_dm_concat":
            hp.randint("fex_dm_concat", 2),
            "fex_dm":
            hp.randint("fex_dm", 3),
            "fex_dbow_words":
            hp.randint("fex_dbow_words", 2),
        })

        return hyper_space, hyper_choices



# Copyright 2019-2022 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from asreview.models.balance.base import BaseBalance


class SimpleBalance(BaseBalance):
    """Simple (no balancing) balance strategy.

    Use all training data.
    """

    name = "simple"
    label = "Simple (no balancing)"

    def sample(self, X, y, train_idx):
        """
        Function that does not resample the training set.

        Arguments
        ---------
        X: numpy.ndarray
            Complete matrix of all samples.
        y: numpy.ndarray
            Classified results of all samples.

        Returns
        -------
        numpy.ndarray:
            Training samples.
        numpy.ndarray:
            Classification of training samples.
        """
        return X[train_idx], y[train_idx]



# Copyright 2019-2022 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
from abc import ABC
from abc import abstractclassmethod
from argparse import RawTextHelpFormatter

from asreview import __version__


class BaseEntryPoint(ABC):
    """Base class for defining entry points."""

    description = "Base Entry point."
    extension_name = "asreview"
    version = __version__

    @abstractclassmethod
    def execute(self, argv):
        """Perform the functionality of the entry point.

        Arguments
        ---------
        argv: list
            Argument list, with the entry point and program removed.
            For example, if `asreview plot X` is executed, then argv == ['X'].
        """
        raise NotImplementedError

    def format(self, entry_name="?"):
        """Create a short formatted description of the entry point.

        Arguments
        ---------
        entry_name: str
            Name of the entry point. For example 'plot' in `asreview plot X`
        """
        description = self.description
        version = getattr(self, "version", "?")
        extension_name = getattr(self, "extension_name", "?")

        display_name = f"{entry_name} [{extension_name}-{version}]"

        return f"{display_name}\n    {description}"


def _base_parser(prog=None, description=None):
    """Argument parser for simulate.

    Parameters
    ----------
    mode : str
        The mode to run ASReview.
    prog : str
        The program name. For example 'asreview'.

    Returns
    -------
    argparse.ArgumentParser
        Configured argparser.
    """

    # parse arguments if available
    parser = argparse.ArgumentParser(
        prog=prog,
        description=description,
        formatter_class=RawTextHelpFormatter
    )
    parser.add_argument(
        "--embedding",
        type=str,
        default=None,
        dest='embedding_fp',
        help="File path of embedding matrix. Required for LSTM models."
    )
    parser.add_argument(
        "--config_file",
        type=str,
        default=None,
        help="Configuration file with model settings"
             "and parameter values."
    )
    parser.add_argument(
        "--seed",
        default=None,
        type=int,
        help="Seed for the model (classifiers, balance "
             "strategies, feature extraction techniques, and query "
             "strategies). Use an integer between 0 and 2^32 - 1."
    )
    return parser



# Copyright 2019-2022 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

ASCII_LOGO = """
            _____ _____            _
     /\    / ____|  __ \          (_)
    /  \  | (___ | |__) |_____   ___  _____      __
   / /\ \  \___ \|  _  // _ \ \ / / |/ _ \ \ /\ / /
  / ____ \ ____) | | \ \  __/\ V /| |  __/\ V  V /
 /_/    \_\_____/|_|  \_\___| \_/ |_|\___| \_/\_/
"""  # noqa

AVAILABLE_CLI_MODI = ["simulate"]
AVAILABLE_REVIEW_CLASSES = ["simulate", "minimal"]

LABEL_NA = -1

KERAS_MODELS = ["lstm_base", "lstm_pool"]

# project types
PROJECT_MODE_ORACLE = "oracle"
PROJECT_MODE_EXPLORE = "explore"
PROJECT_MODE_SIMULATE = "simulate"
PROJECT_MODES = [PROJECT_MODE_ORACLE, PROJECT_MODE_EXPLORE, PROJECT_MODE_SIMULATE]

# CLI defaults
DEFAULT_MODEL = "nb"
DEFAULT_QUERY_STRATEGY = "max"
DEFAULT_BALANCE_STRATEGY = "double"
DEFAULT_FEATURE_EXTRACTION = "tfidf"
DEFAULT_N_INSTANCES = 1
DEFAULT_N_PRIOR_INCLUDED = 1
DEFAULT_N_PRIOR_EXCLUDED = 1

GITHUB_PAGE = "https://github.com/asreview/asreview"
EMAIL_ADDRESS = "asreview@uu.nl"

LEGACY_STATE_EXTENSIONS = [".h5", ".hdf5", ".he5", ".json"]

COLUMN_DEFINITIONS = {
    # included
    "included": [
        "final_included", "label", "label_included", "included_label",
        "included_final", "included", "included_flag", "include"
    ],
    # abstract included (pending deprecation)
    "abstract_included": [
        "abstract_included", "included_abstract", "included_after_abstract",
        "label_abstract_screening"
    ],
    "title": ["title", "primary_title"],
    "authors": ["authors", "author names", "first_authors"],
    "abstract": ["abstract", "abstract note", "notes_abstract"],
    "notes": ["notes"],
    "keywords": ["keywords"],
    "doi": ["doi"],
}

# the schema describes the content of the ASReview project file.
SCHEMA = {
    "$schema": "http://json-schema.org/draft-07/schema",
    "$id": "http://example.com/example.json",
    "type": "object",
    "title": "The ASReview project file root schema",
    "description": "The root schema comprises the entire project.json file in the ASReview project file.",
    "default": {},
    "examples": [
        {
            "version": "1.0",
            "id": "example",
            "mode": "oracle",
            "name": "example",
            "description": "",
            "authors": "",
            "created_at_unix": 1648205610,
            "datetimeCreated": "2022-03-25 11:53:30.510461",
            "reviews": [
                {
                    "id": "4793de70a8d44eb4baa68bac2853c91a",
                    "start_time": "2022-03-25 11:55:50.551360",
                    "status": "review",
                    "end_time": "2022-03-26 10:31:52.441360"
                }
            ],
            "feature_matrices": [
                {
                    "id": "tfidf",
                    "filename": "tfidf_feature_matrix.npz"
                }
            ],
            "dataset_path": "example.ris"
        }
    ],
    "required": [
        "version",
        "id",
        "mode",
        "name"
    ],
    "properties": {
        "version": {
            "$id": "#/properties/version",
            "type": "string",
            "title": "The version schema",
            "description": "The version of ASReview on initiation of the project.",
            "default": "",
            "examples": [
                "1.0"
            ]
        },
        "id": {
            "$id": "#/properties/id",
            "type": "string",
            "title": "The id schema",
            "description": "The unique identifier of the project.",
            "default": "",
            "examples": [
                "example"
            ]
        },
        "mode": {
            "$id": "#/properties/mode",
            "type": "string",
            "title": "The mode schema",
            "description": "The mode of the project. One of oracle, explore, or simulate.",
            "default": "",
            "enum": PROJECT_MODES,
            "examples": [
                "oracle"
            ]
        },
        "name": {
            "$id": "#/properties/name",
            "type": [
                "string",
                "null"
            ],
            "title": "The name schema",
            "description": "The name of the project.",
            "default": "",
            "examples": [
                "example"
            ]
        },
        "description": {
            "$id": "#/properties/description",
            "type": [
                "string",
                "null"
            ],
            "title": "The description schema",
            "description": "The description of the project.",
            "default": "",
            "examples": [
                ""
            ]
        },
        "authors": {
            "$id": "#/properties/authors",
            "type": [
                "string",
                "null"
            ],
            "title": "The authors schema",
            "description": "The authors of the project.",
            "default": "",
            "examples": [
                ""
            ]
        },
        "created_at_unix": {
            "$id": "#/properties/created_at_unix",
            "type": [
                "integer",
                "null"
            ],
            "title": "The created_at_unix schema",
            "description": "An explanation about the purpose of this instance.",
            "default": 0,
            "examples": [
                1648205610
            ]
        },
        "datetimeCreated": {
            "$id": "#/properties/datetimeCreated",
            "type": [
                "string",
                "null"
            ],
            "title": "The datetimeCreated schema",
            "description": "The date and time of the project creation.",
            "default": "",
            "examples": [
                "2022-03-25 11:53:30.510461"
            ]
        },
        "reviews": {
            "$id": "#/properties/reviews",
            "type": "array",
            "title": "The reviews schema",
            "description": "The list of reviews in the project. Multiple reviews per project are possible, however this is limited to 1 at the moment.",
            "default": [],
            "examples": [
                [
                    {
                        "id": "4793de70a8d44eb4baa68bac2853c91a",
                        "start_time": "2022-03-25 11:55:50.551360",
                        "status": "review"
                    }
                ]
            ],
            "additionalItems": True,
            "items": {
                "$id": "#/properties/reviews/items",
                "anyOf": [
                    {
                        "$id": "#/properties/reviews/items/anyOf/0",
                        "type": "object",
                        "title": "The first anyOf schema",
                        "description": "An explanation about the purpose of this instance.",
                        "default": {},
                        "examples": [
                            {
                                "id": "4793de70a8d44eb4baa68bac2853c91a",
                                "start_time": "2022-03-25 11:55:50.551360",
                                "status": "review"
                            }
                        ],
                        "required": [
                            "id",
                            "start_time",
                            "status"
                        ],
                        "properties": {
                            "id": {
                                "$id": "#/properties/reviews/items/anyOf/0/properties/id",
                                "type": "string",
                                "title": "The id of the review.",
                                "description": "A unique UUID4 identifier of the review.",
                                "default": "",
                                "examples": [
                                    "4793de70a8d44eb4baa68bac2853c91a"
                                ]
                            },
                            "start_time": {
                                "$id": "#/properties/reviews/items/anyOf/0/properties/start_time",
                                "type": "string",
                                "title": "The start_time of the review.",
                                "description": "The start date and time of the review.",
                                "default": "",
                                "examples": [
                                    "2022-03-25 11:55:50.551360"
                                ]
                            },
                            "end_time": {
                                "$id": "#/properties/reviews/items/anyOf/0/properties/start_time",
                                "type": "string",
                                "title": "The end_time of the review.",
                                "description": "The end date and time of the review.",
                                "default": "",
                                "examples": [
                                    "2022-03-26 10:31:52.441360"
                                ]
                            },
                            "status": {
                                "$id": "#/properties/reviews/items/anyOf/0/properties/status",
                                "type": [
                                    "string",
                                    "null"
                                ],
                                "title": "The status of the review.",
                                "description": "The status of the review. Options are setup, review, finished.",
                                "enum": ["setup", "review", "finished"],
                                "default": "setup",
                                "examples": [
                                    "review"
                                ]
                            }
                        },
                        "additionalProperties": True
                    }
                ]
            }
        },
        "feature_matrices": {
            "$id": "#/properties/feature_matrices",
            "type": "array",
            "title": "The feature_matrices schema",
            "description": "Information about the feature matrices.",
            "default": [],
            "examples": [
                [
                    {
                        "id": "tfidf",
                        "filename": "tfidf_feature_matrix.npz"
                    }
                ]
            ],
            "additionalItems": True,
            "items": {
                "$id": "#/properties/feature_matrices/items",
                "anyOf": [
                    {
                        "$id": "#/properties/feature_matrices/items/anyOf/0",
                        "type": "object",
                        "title": "The first anyOf schema",
                        "description": "Information about a feature matrix.",
                        "default": {},
                        "examples": [
                            {
                                "id": "tfidf",
                                "filename": "tfidf_feature_matrix.npz"
                            }
                        ],
                        "required": [
                            "id",
                            "filename"
                        ],
                        "properties": {
                            "id": {
                                "$id": "#/properties/feature_matrices/items/anyOf/0/properties/id",
                                "type": "string",
                                "title": "The id schema",
                                "description": "A unique id of the feature matrix.",
                                "default": "",
                                "examples": [
                                    "tfidf"
                                ]
                            },
                            "filename": {
                                "$id": "#/properties/feature_matrices/items/anyOf/0/properties/filename",
                                "type": "string",
                                "title": "The filename schema",
                                "description": "The name of the file with the feature matrix. Usually a sparse matrix.",
                                "default": "",
                                "examples": [
                                    "tfidf_feature_matrix.npz"
                                ]
                            }
                        },
                        "additionalProperties": True
                    }
                ]
            }
        },
        "dataset_path": {
            "$id": "#/properties/dataset_path",
            "type": [
                "string",
                "null"
            ],
            "title": "The dataset_path schema",
            "description": "Name of the dataset file.",
            "default": "",
            "examples": [
                "example.ris"
            ]
        }
    },
    "additionalProperties": True
}



# Copyright 2019-2022 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from asreview.models.query.base import BaseQueryStrategy
from asreview.models.query.utils import get_query_model
from asreview.utils import get_random_state


def _parse_mixed_kwargs(kwargs, strategy_name):

    kwargs_new = {}
    for key, value in kwargs.items():
        if key.startswith(strategy_name):
            new_key = key[len(strategy_name) + 1:]
            kwargs_new[new_key] = value

    return kwargs_new


class MixedQuery(BaseQueryStrategy):
    """Mixed query strategy.

    Use two different query strategies at the same time with a
    ratio of one to the other. A mix of two query strategies is used. For
    example mixing max and random sampling with a mix ratio of 0.95 would mean
    that at each query 95% of the instances would be sampled with the max
    query strategy after which the remaining 5% would be sampled with the
    random query strategy. It would be called the `max_random` query strategy.
    Every combination of primitive query strategy is possible.

    Arguments
    ---------
    strategy_1: str
        Name of the first query strategy. Default 'max'.
    strategy_2: str
        Name of the second query strategy. Default 'random'
    mix_ratio: float
        Sampling from strategy_1 and strategy_2 according a Bernoulli
        distribution. E.g. for mix_ratio=0.95, this implies strategy_1
        with probability 0.95 and strategy_2 with probability 0.05.
        Default 0.95.
    random_state: float
        Seed for the numpy random number generator.
    **kwargs: dict
        Keyword arguments for the two strategy. To specify which of the
        strategies the argument is for, prepend with the name of the query
        strategy and an underscore, e.g. 'max' for maximal sampling.
    """

    def __init__(self,
                 strategy_1="max",
                 strategy_2="random",
                 mix_ratio=0.95,
                 random_state=None,
                 **kwargs):
        """Initialize the Mixed query strategy."""
        super(MixedQuery, self).__init__()

        self.strategy_1 = strategy_1
        self.strategy_2 = strategy_2

        self.mix_ratio = mix_ratio
        self._random_state = get_random_state(random_state)

        self.kwargs_1 = _parse_mixed_kwargs(kwargs, strategy_1)
        self.kwargs_2 = _parse_mixed_kwargs(kwargs, strategy_2)

        self.query_model1 = get_query_model(strategy_1, **self.kwargs_1)
        if "random_state" in self.query_model1.default_param:
            self.query_model1 = get_query_model(
                strategy_1, random_state=self._random_state, **self.kwargs_1)

        self.query_model2 = get_query_model(strategy_2, **self.kwargs_2)
        if "random_state" in self.query_model2.default_param:
            self.query_model2 = get_query_model(
                strategy_2, random_state=self._random_state, **self.kwargs_2)

    def query(self, X, classifier, n_instances=None, **kwargs):

        # set the number of instances to len(X) if None
        if n_instances is None:
            n_instances = X.shape[0]

        # compute the predictions
        predictions = classifier.predict_proba(X)

        # Perform the query with strategy 1.
        try:
            query_idx_1 = self.query_model1._query(
                predictions,
                n_instances=n_instances)
        except AttributeError:
            # for random for example
            query_idx_1 = self.query_model1.query(
                X, classifier, n_instances)

        # Perform the query with strategy 2.
        try:
            query_idx_2 = self.query_model2._query(
                predictions,
                n_instances=n_instances)
        except AttributeError:
            # for random for example
            query_idx_2 = self.query_model2.query(
                X, classifier, n_instances)

        # mix the 2 query strategies into one list
        query_idx_mix = []
        i = 0
        j = 0

        while i < len(query_idx_1) and j < len(query_idx_2):

            if self._random_state.rand() < self.mix_ratio:
                query_idx_mix.append(query_idx_1[i])
                i = i + 1
            else:
                query_idx_mix.append(query_idx_2[j])
                j = j + 1

        return np.unique(query_idx_mix)[0:n_instances]

    def full_hyper_space(self):
        from hyperopt import hp

        space_1, choices_1 = self.query_model1.hyper_space()
        space_2, choices_2 = self.query_model2.hyper_space()
        parameter_space = {}
        hyper_choices = {}
        for key, value in space_1.items():
            new_key = "qry_" + self.strategy_1 + key[4:]
            parameter_space[new_key] = value
            hyper_choices[new_key] = choices_1[key]

        for key, value in space_2.items():
            new_key = "qry_" + self.strategy_2 + key[4:]
            parameter_space[new_key] = value
            hyper_choices[new_key] = choices_2[key]

        parameter_space["qry_mix_ratio"] = hp.uniform("qry_mix_ratio", 0, 1)

        return parameter_space, hyper_choices

    @property
    def name(self):
        return "_".join([self.strategy_1, self.strategy_2])


class MaxRandomQuery(MixedQuery):
    """Mixed (95% Maximum and 5% Random) query strategy.

    A mix of maximum and random query strategies with a mix ratio of 0.95.
    At each query 95% of the instances would be sampled with the maximum
    query strategy after which the remaining 5% would be sampled with
    the random query strategy.
    """

    name = "max_random"
    label = "Mixed (95% Maximum and 5% Random)"

    def __init__(self,
                 mix_ratio=0.95,
                 random_state=None,
                 **kwargs):
        """Initialize the Mixed (Maximum and Random) query strategy."""
        super(MaxRandomQuery, self).__init__(
            strategy_1="max",
            strategy_2="random",
            mix_ratio=mix_ratio,
            random_state=random_state,
            **kwargs)


class MaxUncertaintyQuery(MixedQuery):
    """Mixed (95% Maximum and 5% Uncertainty) query strategy.

    A mix of maximum and random query strategies with a mix ratio of 0.95.
    At each query 95% of the instances would be sampled with the maximum
    query strategy after which the remaining 5% would be sampled with
    the uncertainty query strategy.
    """

    name = "max_uncertainty"
    label = "Mixed (95% Maximum and 5% Uncertainty)"

    def __init__(self,
                 mix_ratio=0.95,
                 random_state=None,
                 **kwargs):
        """Initialize the Mixed (Maximum and Uncertainty) query strategy."""
        super(MaxUncertaintyQuery, self).__init__(
            strategy_1="max",
            strategy_2="uncertainty",
            mix_ratio=mix_ratio,
            random_state=random_state,
            **kwargs)
